@@ -29,8 +29,8 @@ Input, recurrent network, and output layer.
 
 __author__ = "Richard Gast"
 __status__ = "Development"
-__version__ = "0.5.1"
+__version__ = "0.9.3"
 
-from .network import Network
-from .utility import random_connectivity, circular_connectivity, input_connections
+from .network import Network, Observer
+from .utility import random_connectivity, circular_connectivity, input_connections, normalize
 from .utility import wta_score, readout
