@@ -6,7 +6,7 @@ from .input_layer import InputLayer
 from .output_layer import OutputLayer
 from .utility import retrieve_from_dict, add_op_name
 from .observer import Observer
-from pyrates import NodeTemplate
+from pyrates import NodeTemplate, CircuitTemplate
 import numpy as np
 from time import perf_counter
 
@@ -16,7 +16,7 @@ class Network:
     output layers.
     """
 
-    def __init__(self, n: int, rnn_layer: Union[RNNLayer, SRNNLayer], var_map: dict = None):
+    def __init__(self, n: int, rnn_layer: Union[RNNLayer, SRNNLayer], var_map: dict = None, device: str = "cpu"):
         """Instantiates network with a single RNN layer.
 
         Parameters
@@ -28,6 +28,9 @@ class Network:
         var_map
             Optional dictionary, where keys are variable names as supplied by the user, and values are variable names
             that contain the operator name as well.
+        device
+            Device on which to deploy the `Network` instance.
+
         """
 
         self.n = n
@@ -36,13 +39,18 @@ class Network:
         self.output_layer = None
         self._var_map = var_map if var_map else {}
         self._model = None
-
-    def __getitem__(self, item: int):
+        self.device = device
+        
+    def __getitem__(self, item: Union[int, str]):
         try:
-            return self._model[item]
-        except TypeError:
-            self.compile()
-        return self._model[item]
+            return self.rnn_layer[item]
+        except KeyError:
+            try:
+                return self.rnn_layer[self._var_map[item]]
+            except KeyError:
+                if self._model is None:
+                    self.compile()
+                return self._model[item]
 
     def __len__(self):
         try:
@@ -54,7 +62,7 @@ class Network:
     @classmethod
     def from_yaml(cls, node: Union[str, NodeTemplate], weights: np.ndarray, source_var: str, target_var: str,
                   input_var: str, output_var: str, spike_var: str = None, spike_def: str = None, op: str = None,
-                  train_params: list = None, record_vars: list = None, **kwargs):
+                  train_params: list = None, device: str = "cpu", **kwargs):
         """Creates a `Network` instance from a YAML template that defines a single RNN node and additional information
         about which nodes in the network should be connected to each other.
 
@@ -84,8 +92,8 @@ class Network:
             the operator name is provided together with the variable names, e.g. `source_var = <op>/<var>`.
         train_params
             Names of all RNN parameters that should be made available for optimization.
-        record_vars
-            Names of all variables that should be made available for recordings during training/testing/simulation runs.
+        device
+            Device on which to deploy the `Network` instance.
         kwargs
             Additional keyword arguments provided to the `RNNLayer` (or `SRNNLayer` in case of spiking neurons).
 
@@ -104,25 +112,98 @@ class Network:
                 var_dict[key] = add_op_name(op, var, new_vars)
             if train_params:
                 train_params = [add_op_name(op, p, new_vars) for p in train_params]
-            if record_vars:
-                record_vars = [add_op_name(op, v, new_vars) for v in record_vars]
+            if "node_vars" in kwargs:
+                for key in kwargs["node_vars"].copy():
+                    if "/" not in key:
+                        val = kwargs["node_vars"].pop(key)
+                        kwargs["node_vars"][f"all/{op}/{key}"] = val
 
         # initialize rnn layer
         if spike_var is None and spike_def is None:
             rnn_layer = RNNLayer.from_yaml(node, weights, var_dict['svar'], var_dict['tvar'], var_dict['in_ext'],
-                                           var_dict['out'], train_params=train_params, record_vars=record_vars,
-                                           **kwargs)
+                                           var_dict['out'], train_params=train_params, device=device, **kwargs)
         elif spike_var is None or spike_def is None:
             raise ValueError('To define a reservoir with a spiking neural network layer, please provide both the '
-                             'name of the variable that spikes should be stored in (`spike_def`) as well as the '
+                             'name of the variable that spikes should be stored in (`spike_var`) as well as the '
                              'name of the variable that is used to define spikes (`spike_def`).')
         else:
             rnn_layer = SRNNLayer.from_yaml(node, weights, var_dict['svar'], var_dict['tvar'], var_dict['in_ext'],
                                             var_dict['out'], spike_def=var_dict['spike'], spike_var=var_dict['in_net'],
-                                            train_params=train_params, record_vars=record_vars, **kwargs)
+                                            train_params=train_params, device=device, **kwargs)
+
+        # remember operator mapping for each RNN layer parameter and state variable
+        for p in rnn_layer.parameter_names:
+            add_op_name(op, p, new_vars)
+        for v in rnn_layer.variable_names:
+            add_op_name(op, v, new_vars)
 
         # initialize model
-        return cls(weights.shape[0], rnn_layer, var_map=new_vars)
+        return cls(weights.shape[0], rnn_layer, var_map=new_vars, device=device)
+
+    @classmethod
+    def from_template(cls, template: CircuitTemplate, input_var: str, output_var: str, spike_var: str = None,
+                      spike_def: str = None, op: str = None, train_params: list = None, device: str = "cpu", **kwargs):
+        """Creates a `Network` instance from a YAML template that defines a single RNN node and additional information
+        about which nodes in the network should be connected to each other.
+
+        Parameters
+        ----------
+        template
+            Instance of a `pyrates.CircuitTemplate`. Will not be altered any further.
+        input_var
+            Name of the parameter in the node equations that input from input layers should be projected to.
+        output_var
+            Name of the variable in the node equations that should be used as output of the RNN layer.
+        spike_var
+            Name of the parameter in the node equations that recurrent input from the RNN layer should be projected to.
+        spike_def
+            Name of the variable in the node equations that should be used to determine spikes in the network.
+        op
+            Name of the operator in which all the above variables can be found. If not provided, it is assumed that
+            the operator name is provided together with the variable names, e.g. `source_var = <op>/<var>`.
+        train_params
+            Names of all RNN parameters that should be made available for optimization.
+        device
+            Device on which to deploy the `Network` instance.
+        kwargs
+            Additional keyword arguments provided to the `RNNLayer` (or `SRNNLayer` in case of spiking neurons).
+
+        Returns
+        -------
+        Network
+            Instance of `Network`.
+        """
+
+        # add operator key to variable names
+        var_dict = {'in_ext': input_var, 'in_net': spike_var, 'out': output_var, 'spike': spike_def}
+        new_vars = {}
+        if op is not None:
+            for key, var in var_dict.copy().items():
+                var_dict[key] = add_op_name(op, var, new_vars)
+            if train_params:
+                train_params = [add_op_name(op, p, new_vars) for p in train_params]
+
+        # initialize rnn layer
+        if spike_var is None and spike_def is None:
+            rnn_layer = RNNLayer.from_template(template, var_dict['in_ext'], var_dict['out'], train_params=train_params,
+                                               **kwargs)
+        elif spike_var is None or spike_def is None:
+            raise ValueError('To define a reservoir with a spiking neural network layer, please provide both the '
+                             'name of the variable that spikes should be stored in (`spike_var`) as well as the '
+                             'name of the variable that is used to define spikes (`spike_def`).')
+        else:
+            rnn_layer = SRNNLayer.from_template(template, var_dict['in_ext'], var_dict['out'],
+                                                spike_def=var_dict['spike'], spike_var=var_dict['in_net'],
+                                                train_params=train_params, **kwargs)
+
+        # remember operator mapping for each RNN layer parameter and state variable
+        for p in rnn_layer.parameter_names:
+            add_op_name(op, p, new_vars)
+        for v in rnn_layer.variable_names:
+            add_op_name(op, v, new_vars)
+
+        # initialize model
+        return cls(len(template.nodes), rnn_layer, var_map=new_vars, device=device)
 
     @property
     def model(self) -> Sequential:
@@ -157,7 +238,7 @@ class Network:
         input_layer = InputLayer(self.n, m, weights, trainable=trainable, dtype=dtype)
 
         # add layer to model
-        self.input_layer = input_layer
+        self.input_layer = input_layer.to(self.device)
 
         # return layer
         return self.input_layer
@@ -176,7 +257,12 @@ class Network:
         trainable
             If true, the output weights will be made available for optimization.
         activation_function
-            Optional activation function applied to the output of the output layer.
+            Optional activation function applied to the output of the output layer. Valid options are:
+            - 'tanh' for `torch.nn.Tanh()`
+            - 'sigmoid' for `torch.nn.Sigmoid()`
+            - 'softmax' for `torch.nn.Softmax(dim=0)`
+            - `softmin` for `torch.nn.Softmin(dim=0)`
+            - None (default) for `torch.nn.Identity()`
         dtype
             Data type of the input weights.
         kwargs
@@ -193,7 +279,7 @@ class Network:
                                    dtype=dtype, **kwargs)
 
         # add layer to model
-        self.output_layer = output_layer
+        self.output_layer = output_layer.to(self.device)
 
         # return layer
         return self.output_layer
@@ -209,8 +295,8 @@ class Network:
         self.output_layer = None
 
     def train(self, inputs: np.ndarray, targets: np.ndarray, optimizer: str = 'sgd', optimizer_kwargs: dict = None,
-              loss: str = 'mse', loss_kwargs: dict = None, lr: float = 1e-3, device: str = None,
-              sampling_steps: int = 100, optimizer_steps: int = 1, verbose: bool = True, **kwargs) -> Observer:
+              loss: str = 'mse', loss_kwargs: dict = None, lr: float = 1e-3, sampling_steps: int = 100,
+              optimizer_steps: int = 1, verbose: bool = True, **kwargs) -> Observer:
         """Optimize model parameters such that the model output matches the provided targets as close as possible.
 
         Parameters
@@ -244,10 +330,11 @@ class Network:
             Additional keyword arguments provided to the initialization of the loss.
         lr
             Learning rate.
-        device
-            Device on which to deploy the network model.
         sampling_steps
             Number of training steps at which to record observables.
+        optimizer_steps
+            Number of training steps after which to perform an update of the trainable parameters based on the
+            accumulated gradients.
         verbose
             If true, the training progress will be displayed.
         kwargs
@@ -270,7 +357,7 @@ class Network:
                              '`targets` agree in the first dimension.')
 
         # set up model
-        model = self.compile(device) if self._model is None else self._model
+        model = self.compile() if self._model is None else self._model
 
         # initialize loss function
         loss = self._get_loss_function(loss, loss_kwargs=loss_kwargs)
@@ -302,7 +389,7 @@ class Network:
         return obs
 
     def test(self, inputs: np.ndarray, targets: np.ndarray, loss: str = 'mse', loss_kwargs: dict = None,
-             device: str = None, sampling_steps: int = 100, verbose: bool = True, **kwargs) -> tuple:
+             sampling_steps: int = 100, verbose: bool = True, **kwargs) -> tuple:
         """Test the model performance on a set of inputs and target outputs, with frozen model parameters.
 
         Parameters
@@ -318,8 +405,6 @@ class Network:
             for available options.
         loss_kwargs
             Additional keyword arguments provided to the initialization of the loss.
-        device
-            Device on which to deploy the network model.
         sampling_steps
             Number of testing steps at which to record observables.
         verbose
@@ -341,7 +426,7 @@ class Network:
         target_tensor = torch.tensor(targets)
 
         # set up model
-        model = self.compile(device) if self._model is None else self._model
+        model = self.compile() if self._model is None else self._model
 
         # initialize loss function
         loss = self._get_loss_function(loss, loss_kwargs=loss_kwargs)
@@ -370,11 +455,11 @@ class Network:
                 if step % sampling_steps == 0:
                     if verbose:
                         print(f'Progress: {step}/{steps} test steps finished.')
-                    obs.record(prediction, error.item(), self.rnn_layer.record(rec_vars))
+                    obs.record(step, prediction, error.item(), [self[v] for v in rec_vars])
 
         return obs, loss_total
 
-    def run(self, inputs: np.ndarray, device: str = 'cpu', sampling_steps: int = 1, verbose: bool = True, **kwargs
+    def run(self, inputs: np.ndarray, sampling_steps: int = 1, verbose: bool = True, **kwargs
             ) -> Observer:
         """Perform numerical integration of the input-driven network equations.
 
@@ -383,8 +468,6 @@ class Network:
         inputs
             `T x m` array of inputs fed to the model, where`T` is the number of integration steps and `m` is the number
              of input dimensions of the network.
-        device
-            Device on which to deploy the network model.
         sampling_steps
             Number of integration steps at which to record observables.
         verbose
@@ -400,13 +483,13 @@ class Network:
 
         # transform input into tensor
         steps = inputs.shape[0]
-        inp_tensor = torch.tensor(inputs)
+        inp_tensor = torch.tensor(inputs, device=self.device)
 
         # initialize model from layers
-        model = self.compile(device) if self._model is None else self._model
+        model = self.compile() if self._model is None else self._model
 
         # initialize observer
-        obs = Observer(dt=self.rnn_layer.dt, record_loss=False, **kwargs)
+        obs = Observer(dt=self.rnn_layer.dt, record_loss=kwargs.pop("record_loss", False), **kwargs)
         rec_vars = [self._relabel_var(v) for v in obs.recorded_rnn_variables]
 
         # forward input through static network
@@ -416,18 +499,16 @@ class Network:
                 if step % sampling_steps == 0:
                     if verbose:
                         print(f'Progress: {step}/{steps} integration steps finished.')
-                    obs.record(output, 0.0, self.rnn_layer.record(rec_vars))
+                    obs.record(step, output, 0.0, [self[v] for v in rec_vars])
 
         return obs
 
-    def compile(self, device: str = None) -> Sequential:
+    def compile(self) -> Sequential:
         """Connects the `InputLayer`, `RNNLayer` and `OutputLayer` of this model via a `torch.nn.Sequential` instance.
         Input and output layers are optional.
 
         Parameters
         ----------
-        device
-            Device on which to deploy the `torch.nn.Sequential` instance.
 
         Returns
         -------
@@ -439,15 +520,13 @@ class Network:
         rnn_layer = self._get_layer(self.rnn_layer)
         layers = in_layer + rnn_layer + out_layer
         model = Sequential(*layers)
-        self._model = model
+        self._model = model.to(self.device)
         if not list(self.rnn_layer.parameters()):
             self.rnn_layer.detach()
         else:
             for p in self.parameters():
                 p.requires_grad = True
-        if device is not None:
-            model.to(device)
-        return model
+        return self._model
 
     def forward(self, x: torch.Tensor) -> torch.Tensor:
         """Forward method as implemented for any `torch.Module`.
@@ -510,7 +589,7 @@ class Network:
             if step % sampling_steps == 0:
                 if verbose:
                     print(f'Progress: {step}/{steps} training steps finished.')
-                obs.record(prediction, error.item(), self.rnn_layer.record(rec_vars))
+                obs.record(step, prediction, error.item(), [self[v] for v in rec_vars])
 
     def _train_epochs(self, inp: torch.Tensor, target: torch.Tensor, model: Sequential, loss: Callable,
                       optimizer: torch.optim.Optimizer, obs: Observer, rec_vars: list, error_kwargs: dict,
@@ -519,6 +598,7 @@ class Network:
         if inp.shape[0] != target.shape[0] or inp.shape[1] != target.shape[1]:
             raise ValueError('Wrong dimensions of input and target output. Please make sure that `inputs` and '
                              '`targets` agree in the first dimension (epochs) and second dimension (steps per epoch).')
+
         epochs = inp.shape[0]
         steps = inp.shape[1]
         for epoch in range(epochs):
@@ -542,7 +622,7 @@ class Network:
 
                 # results storage
                 if step % sampling_steps == 0:
-                    obs.record(prediction, error.item(), self.rnn_layer.record(rec_vars))
+                    obs.record(step, prediction, error.item(), [self[v] for v in rec_vars])
 
             # display progress
             if verbose:
