@@ -1,4 +1,5 @@
 import numpy as np
+import torch
 from scipy.stats import rv_discrete, bernoulli
 from typing import Union
 
@@ -46,8 +47,12 @@ def add_op_name(op: str, var: Union[str, None], new_var_names: dict) -> Union[st
     str
         Updated variable name.
     """
-    if var is None or '/' in var:
+    if var is None:
         return var
+    elif "/" in var:
+        _, v = var.split("/")
+        new_var_names[v] = var
+        return new_var_names[v]
     new_var_names[var] = f"{op}/{var}"
     return new_var_names[var]
 
@@ -58,6 +63,13 @@ def _wrap(idxs: np.ndarray, N: int) -> np.ndarray:
     return idxs
 
 
+def to_device(x: torch.Tensor, device: str) -> torch.Tensor:
+    try:
+        return x.to(device)
+    except AttributeError:
+        return x
+
+
 # connectivity generation functions
 ###################################
 
@@ -146,10 +158,52 @@ def input_connections(n: int, m: int, p: float, variance: float = 1.0, zero_mean
         rows = np.argwhere(C_tmp[:, col] > 0).squeeze()
         C[rows, col] = np.random.randn(rows.shape[0])*variance
         if zero_mean:
-            C[rows, col] -= np.sum(C[rows, col])
+            C[rows, col] -= np.sum(C[:, col])/len(rows)
     return C
 
 
+def normalize(x: np.ndarray, mode: str = "minmax", row_wise: bool = False) -> np.ndarray:
+    """Normalization function for matrices.
+
+    Parameters
+    ----------
+    x
+        N x m matrix.
+    mode
+        Normalization mode. Can be one of the following options:
+        - 'minmax': Normalize such that the minimum of the data is 0 and the maximum is 1.
+        - 'zscore': Normalize data such that the mean is 0 and the standard deviation is 1.
+        - 'sum': Normalize such that the sum over the data equals 1.
+    row_wise
+        If true, normalization will be applied independently for each row of `x`.
+
+    Returns
+    -------
+    np.ndarray
+        N x m matrix, normalized.
+    """
+    if row_wise:
+        for i in range(x.shape[0]):
+            x[i, :] = normalize(x[i, :], mode=mode, row_wise=False)
+    else:
+        x_tmp = x.flatten()
+        if mode == "minmax":
+            x -= np.min(x_tmp)
+            max_val = np.max(x_tmp)
+            if max_val > 0:
+                x /= max_val
+        elif mode == "zscore":
+            x -= np.mean(x_tmp)
+            std = np.std(x_tmp)
+            if std > 0:
+                x /= std
+        elif mode == "sum":
+            x /= np.sum(x_tmp)
+        else:
+            raise ValueError(f"Invalid normalization mode: {mode}.")
+    return x
+
+
 # function for optimization
 ###########################
 
@@ -176,7 +230,8 @@ def wta_score(x: np.ndarray, y: np.ndarray) -> float:
     return float(np.mean(z))
 
 
-def readout(X: np.ndarray, y: np.ndarray, k: int = 1, verbose: bool = True, **kwargs) -> tuple:
+def readout(X: np.ndarray, y: np.ndarray, k: int = 1, test_size: float = 0.2, verbose: bool = True,
+            normalize: str = None, **kwargs) -> dict:
     """Uses Ridge regression to find a set of coefficients `a` that minimizes `y - aX`.
 
     Parameters
@@ -184,49 +239,100 @@ def readout(X: np.ndarray, y: np.ndarray, k: int = 1, verbose: bool = True, **kw
     X
         2D array of data (rows = samples, columns = features).
     y
-        2D array of data (rows = samples, columns = output dimensions).
+        2D array of data (rows = samples, columns = output dimensions) or 1D array of samples.
     k
         If larger 1, `k` splits into training and testing data will be performed, and for each of these splits a ridge
         regression will be calculated.
+    test_size
+        Fraction of the input data that is used for independent testing after the training is done.
     verbose
         If true, updates about the regression procedure will be displayed.
+    method
+        Regression method/model to be used. Can be one of 'Ridge', 'RidgeClassifier', 'LogisticRegression',
+        'ElasticNet', or 'Lasso'.
+    normalize
+        Choose from "minmax" or "standard" to scale the input data.
     kwargs
-        Additional keyword arguments passed to `sklearn.linear_model.Ridge`
+        Additional keyword arguments passed to `sklearn.linear_model.SGDClassifier`
 
     Returns
     -------
-    tuple
-        Average loss on training data, and readout weights.
+    dict
+        Contains entries with the following keys:
+         - 'train_score': R2 score training data
+         - 'test_score': R2 score on test data
+         - 'readout_weights': Regression coefficients
+         - 'readout_bias': Regression bias term
+         - 'prediction': Prediction of the fitted classifier for the testing data
+         - 'target': Prediction target, i.e. the testing data
     """
 
     # imports
-    from sklearn.linear_model import Ridge
-    from sklearn.model_selection import StratifiedKFold
+    from sklearn.linear_model import SGDClassifier
+    from sklearn.model_selection import StratifiedKFold, train_test_split
+    from sklearn.preprocessing import scale, minmax_scale
+
+    # initialize classifier
+    classifier = SGDClassifier(**kwargs)
+
+    # normalize input data
+    if normalize == "minmax":
+        X = minmax_scale(X)
+    elif normalize == "standard":
+        X = scale(X)
+
+    # split into training and test data set
+    if test_size > 0.0:
+        X, X_t, y, y_t = train_test_split(X, y, test_size=test_size, shuffle=False)
 
     # perform ridge regression
     if k > 1:
         splitter = StratifiedKFold(n_splits=k)
-        scores, coefs = [], []
+        scores, coefs, intercepts = [], [], []
         for i, (train_idx, test_idx) in enumerate(splitter.split(X=X, y=y)):
-            classifier = Ridge(**kwargs)
             classifier.fit(X[train_idx], y[train_idx])
             scores.append(classifier.score(X=X[test_idx], y=y[test_idx]))
             coefs.append(classifier.coef_)
+            intercepts.append(classifier.intercept_)
     else:
-        classifier = Ridge(**kwargs)
         classifier.fit(X, y)
         scores = [classifier.score(X=X, y=y)]
         coefs = [classifier.coef_]
+        intercepts = [classifier.intercept_]
 
     # store readout weights
     w_out = np.mean(coefs, axis=0)
-    avg_score = np.mean(scores)
+    intercept = np.mean(intercepts, axis=0)
+    train_score = np.mean(scores)
 
     if verbose:
         print(f'Finished readout training.')
         if k > 1:
-            print(fr'Average, cross-validated $R^2$ score across {k} test folds: {avg_score}')
+            print(fr'Average, cross-validated score across {k} test folds: {train_score}')
         else:
-            print(fr'$R^2$ score on training data: {avg_score}')
+            print(fr'Score on training data: {train_score}')
 
-    return avg_score, w_out
+    # store temporary results
+    res = {"train_score": train_score, "readout_weights": w_out, "readout_bias": intercept, "test_score": None,
+           "prediction": None, "target": None}
+
+    # perform testing
+    if test_size > 0.0:
+
+        classifier.coef_ = w_out
+        classifier.intercept_ = intercept
+
+        res["test_score"] = classifier.score(X=X_t, y=y_t)
+        res["prediction"] = classifier.predict(X_t)
+        res["target"] = y_t
+
+        if verbose:
+            print(f'Finished readout training.')
+            print(fr"Score on testing data: {res['test_score']}")
+
+    else:
+
+        res["prediction"] = classifier.predict(X)
+        res["target"] = y
+
+    return res
